/* --------------------------------------------------------------------------
$Id: vhlhdf.c,v 1.8 2001/01/18 12:39:32 ahenja Exp ahenja $
Program:        

Description:    Common functions for working with the HighLevel
		protocol.

Author(s):      Anders Henja

Copyright:	Swedish Meteorological and Hydrological Institute, 2000

History:
$Log: vhlhdf.c,v $
 * Revision 1.8  2001/01/18  12:39:32  ahenja
 * *** empty log message ***
 *
Revision 1.7  2000/03/22 08:39:38  ahenja
Updated with spell checking and header information.

-----------------------------------------------------------------------------*/
#include "write_hlhdf.h"
#include "read_hlhdf.h"
#include "vhlhdf.h"
#include "read_vhlhdf.h"

HL_Node* newHL_Node(const char* name)
{
   HL_Node* retv=NULL;

   if(!name) {
      if(_debug_hdf)
	 fprintf(stderr,"When creating a nodelist item, name has to be specified\n");
   }

   if(!(retv=malloc(sizeof(HL_Node)))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate HL_Node\n");
      return NULL;
   }
   retv->type=UNDEFINED_ID;
   strcpy(retv->name,name);
   retv->ndims=0;
   memset(retv->dims,0,sizeof(size_t)*4);
   retv->data=NULL;
   retv->rawdata = NULL; /* Ahe added this 2001-01-17 */
   strcpy(retv->format,DATAFORMAT_UNDEFINED);
   retv->typeId=-1;
   retv->dSize=0;
   retv->rdSize = 0;
   retv->dataType=DTYPE_UNDEFINED_ID;
   retv->hdfId=-1;
   retv->mark=NMARK_CREATED;
   retv->compoundDescription=NULL;
   return retv;
}

HL_NodeList* newHL_NodeList()
{
   HL_NodeList* retv=NULL;
   int i;

   if(!(retv=malloc(sizeof(HL_NodeList)))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate memory for NODE\n");
      return NULL;
   }
   strcpy(retv->filename,"");

   if(!(retv->nodes = malloc(sizeof(HL_Node*)*DEFAULT_SIZE_NODELIST))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate memory for HL_NodeList\n");
      free(retv);
      return NULL;
   }
   for(i=0;i<DEFAULT_SIZE_NODELIST;i++) {
      retv->nodes[i]=NULL;
   }
   retv->nNodes=0;
   retv->nAllocNodes=DEFAULT_SIZE_NODELIST;
   strcpy(retv->tmp_name,"");
   return retv;
}

void freeHL_Node(HL_Node* node)
{
   H5I_type_t tid;

   if(!node)
      return;

   if(node->typeId>=0) {
      disableErrorReporting();
      H5Tclose(node->typeId);
      enableErrorReporting();
   }

   if(node->hdfId>=0) {
      switch(node->type) {
      case ATTRIBUTE_ID:
	 H5Aclose(node->hdfId);
	 break;
      case DATASET_ID:
	 H5Dclose(node->hdfId);
	 break;
      case GROUP_ID:
	 H5Gclose(node->hdfId);
	 break;
      case TYPE_ID:
	 disableErrorReporting();
	 H5Tclose(node->hdfId);
	 enableErrorReporting();
	 break;
      default:
	 fprintf(stderr,"Strane node type, can't close it (%ld)\n",(long)node->hdfId);
	 break;
      }
   }

   if(node->data)
      free(node->data);
   if(node->rawdata)/* Ahe added this 2001-01-17 */
      free(node->rawdata);
   if(node->compoundDescription)
      freeHL_CompoundTypeDescription(node->compoundDescription);
   free(node);
}

void freeHL_NodeList(HL_NodeList* nodelist)
{
   int i;
   if(!nodelist)
      return;
   if(nodelist->nodes) {
      for(i=0;i<nodelist->nNodes;i++) {
	 freeHL_Node(nodelist->nodes[i]);
      }
      free(nodelist->nodes);
   }
   free(nodelist);

}

HL_Node* newHL_Group(const char* name)
{
   HL_Node* retv=NULL;
   if(!(retv=newHL_Node(name))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate HL_NodeList item group\n");
      return NULL;
   }
   retv->type=GROUP_ID;
   return retv;
}

HL_Node* newHL_Attribute(const char* name)
{
   HL_Node* retv=NULL;
   if(!(retv=newHL_Node(name))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate HL_NodeList item attribute\n");
      return NULL;
   }
   retv->type=ATTRIBUTE_ID;
   return retv;
}

HL_Node* newHL_Dataset(const char* name)
{
   HL_Node* retv=NULL;
   if(!(retv=newHL_Node(name))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate HL_NodeList item attribute\n");
      return NULL;
   }
   retv->type=DATASET_ID;
   return retv;
}

HL_Node* newHL_Datatype(const char* name)
{
   HL_Node* retv=NULL;
   if(!(retv=newHL_Node(name))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate HL_NodeList item attribute\n");
      return NULL;
   }
   retv->type=TYPE_ID;
   return retv;
}

HL_CompoundTypeAttribute* newHL_CompoundTypeAttribute(char* attrname,
						      size_t offset,
						      char* format,
						      size_t size,
						      int ndims,
						      size_t* dims)
{
   HL_CompoundTypeAttribute* retv=NULL;
   int i;
   if(!attrname) {
      if(_debug_hdf)
	 fprintf(stderr,"Impossible to have an attribute without a name in a compound attribute\n");
      return NULL;
   }
   if(!format) {
      if(_debug_hdf)
	 fprintf(stderr,"Impossible to have an attribute without a format in a compound type\n");
      return NULL;
   }
   if(!(retv=malloc(sizeof(HL_CompoundTypeAttribute)))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate CompoundTypeAttribute description\n");
      return NULL;
   }
   strcpy(retv->attrname,attrname);
   retv->offset=offset;
   retv->size=size;
   strcpy(retv->format,format);
   retv->ndims=ndims;
   for(i=0;i<ndims;i++)
      retv->dims[i]=dims[i];

   return retv;
}

HL_CompoundTypeDescription* newHL_CompoundTypeDescription()
{
   HL_CompoundTypeDescription* retv=NULL;
   int i;
   if(!(retv=malloc(sizeof(HL_CompoundTypeDescription)))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate memory for CompoundTypeDescription\n");
      return NULL;
   }
   strcpy(retv->typename,"");
   retv->size=0;
   if(!(retv->attrs = malloc(sizeof(HL_CompoundTypeDescription*)*DEFAULT_SIZE_NODELIST))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate memory for CompoundTypeDescription list\n");
      free(retv);
      return NULL;
   }
   for(i=0;i<DEFAULT_SIZE_NODELIST;i++) {
      retv->attrs[i]=NULL;
   }
   retv->nAttrs=0;
   retv->nAllocAttrs=DEFAULT_SIZE_NODELIST;
   retv->objno[0]=0;
   retv->objno[1]=0;
   return retv;
}

void freeHL_CompoundTypeAttribute(HL_CompoundTypeAttribute* attr)
{
   if(!attr)
      return;
   free(attr);
}

void freeHL_CompoundTypeDescription(HL_CompoundTypeDescription* typelist)
{
   int i;
   if(!typelist)
      return;
   if(typelist->attrs) {
      for(i=0;i<typelist->nAttrs;i++) {
	 if(typelist->attrs[i])
	    freeHL_CompoundTypeAttribute(typelist->attrs[i]);
      }
      free(typelist->attrs);
   }
   free(typelist);
}

int addCompoundTypeAttribute(HL_CompoundTypeDescription* typelist,
			     HL_CompoundTypeAttribute* attribute)
{
   int newallocsize;
   int i;

   if(!attribute) {
      if(_debug_hdf)
	 fprintf(stderr,"Trying to add compound type attribute which is NULL\n");
      return 0;
   }

   if(typelist->nAttrs<typelist->nAllocAttrs-1) {
      typelist->attrs[typelist->nAttrs++]=attribute;
      return 1;
   }

   newallocsize = typelist->nAllocAttrs + DEFAULT_SIZE_NODELIST;
   if(!(typelist->attrs = realloc(typelist->attrs,
				  sizeof(HL_CompoundTypeAttribute*)*newallocsize))) {
      if(_debug_hdf)
	 fprintf(stderr,"Serious memory error occured when reallocating compound attr list\n");
      return 0;
   }
   for(i=typelist->nAllocAttrs;i<newallocsize;i++) {
      typelist->attrs[i]=NULL;
   }
   typelist->nAllocAttrs=newallocsize;
   typelist->attrs[typelist->nAttrs++]=attribute;

   return 1;
}

HL_Node* copyHL_Node(HL_Node* node)
{
   int i,npts;
   HL_Node* retv=NULL;
   if(!node)
      return NULL;
   retv=newHL_Node(node->name);
   retv->type=node->type;
   retv->ndims=node->ndims;
   retv->dSize=node->dSize;
   retv->rdSize = node->rdSize;
   memcpy(retv->dims,node->dims,4*sizeof(hsize_t));
   npts=1;
   for(i=0;i<retv->ndims;i++)
      npts*=retv->dims[i];
   retv->data = malloc(npts*retv->dSize);
   memcpy(retv->data,node->data,npts*retv->dSize);
   
   if(node->rawdata!=NULL) {
      retv->rawdata = malloc(npts*retv->rdSize);
      memcpy(retv->rawdata,node->rawdata,npts*retv->rdSize);
   } else {
      retv->rdSize = 0;
      retv->rawdata = NULL;
   }
   strcpy(retv->format,node->format);

   if(node->typeId>=0)
      retv->typeId=H5Tcopy(node->typeId);

   retv->dataType=node->dataType;
   retv->hdfId=-1;
   retv->mark=node->mark;

   retv->compoundDescription=copyCompoundTypeDescription(node->compoundDescription);

   return retv;
}

HL_CompoundTypeDescription* copyCompoundTypeDescription(HL_CompoundTypeDescription* descr)
{
   /*fprintf(stderr,"copyCompoundTypeDescription not implemented yet!!!\n");*/
   HL_CompoundTypeDescription* retv = NULL;
   int i;
   
   if(descr==NULL)
      return NULL;
   
   if(!(retv=newHL_CompoundTypeDescription())) {
      return NULL;
   }
   strcpy(retv->typename,descr->typename);
   memcpy(retv->objno,descr->objno,sizeof(unsigned long)*2);
   retv->size = descr->size;
   retv->nAttrs = descr->nAttrs;
   retv->nAllocAttrs = descr->nAllocAttrs;

   if(!(retv->attrs = malloc(sizeof(HL_CompoundTypeAttribute*)*retv->nAllocAttrs))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate list of HL_CompoundTypeAttribute\n");
      goto fail;
   }

   for(i=0;i<retv->nAllocAttrs;i++)
      retv->attrs[i]=NULL;
   for(i=0;i<retv->nAttrs;i++) {
      if(!(retv->attrs[i] = newHL_CompoundTypeAttribute(descr->attrs[i]->attrname,
							descr->attrs[i]->offset,
							descr->attrs[i]->format,
							descr->attrs[i]->size,
							descr->attrs[i]->ndims,
							descr->attrs[i]->dims))) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to allocate HL_CompoundTypeAttribute\n");
	 goto fail;
      }
   }
   
   return retv;
fail:
   freeHL_CompoundTypeDescription(retv);
   return NULL;
}

/*
 * If 1 is returned, then responsibility has been taken, otherwise
 * the caller has to free both nodelist and node by him self
 */
int addNode(HL_NodeList* nodelist, HL_Node* node)
{
   int newallocsize;
   int i;
   char tmpStr[256];
   char* tmpPtr;
   int treeStructureOk=0;
   strcpy(tmpStr,node->name);
   

   for(i=0;i<nodelist->nNodes;i++) {
      if(strcmp(nodelist->nodes[i]->name,node->name)==0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Node %s already exists\n",node->name);
	 return 0;
      }
   }

   if(!(tmpPtr=strrchr(tmpStr,'/'))) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not extract '/' from node name %s\n",node->name);
      return 0;
   } else {
      tmpPtr[0]='\0';
      if(strcmp(tmpStr,"")!=0) {
	 for(i=0;treeStructureOk==0&&i<nodelist->nNodes;i++) {
	    if(strcmp(tmpStr,nodelist->nodes[i]->name)==0) {
	       if( (nodelist->nodes[i]->type==GROUP_ID) ||
		   (nodelist->nodes[i]->type==DATASET_ID && node->type==ATTRIBUTE_ID) )
	       treeStructureOk=1;
	    }
	 }
      } else {
	 treeStructureOk=1;
      }
   }

   if(treeStructureOk==0) {
      if(_debug_hdf)
	 fprintf(stderr,"Tree structure not built correct, missing group or dataset %s\n",tmpStr);
      return 0;
   }
   
   if(nodelist->nNodes < nodelist->nAllocNodes-1) {
      nodelist->nodes[nodelist->nNodes++]=node;
      return 1;
   }

   newallocsize = nodelist->nAllocNodes + DEFAULT_SIZE_NODELIST;
   if(!(nodelist->nodes = realloc(nodelist->nodes,sizeof(HL_Node*)*newallocsize))) {
      if(_debug_hdf)
	 fprintf(stderr,"Serious memory error occured when reallocating Node list\n");
      return 0;
   }
   
   for(i=nodelist->nAllocNodes;i<newallocsize;i++) {
      nodelist->nodes[i]=NULL;
   }
   nodelist->nAllocNodes=newallocsize;
   nodelist->nodes[nodelist->nNodes++]=node;

   return 1;
}

HL_Node* getNode(HL_NodeList* nodelist,const char* nodeName)
{
   int i;
   if(!nodelist || !nodeName) {
      if(_debug_hdf)
	 fprintf(stderr,"Can't get node when either nodelist or nodeName is NULL\n");
      return NULL;
   }

   for(i=0;i<nodelist->nNodes;i++) {
      if(strcmp(nodelist->nodes[i]->name,nodeName)==0) {
	 return nodelist->nodes[i];
      }
   }

   if(_debug_hdf)
      fprintf(stderr,"Could not locate node '%s'\n",nodeName);

   return NULL;
}

int setScalarValue(HL_Node* node,size_t sz,unsigned char* value,const char* fmt,hid_t typid)
{
   if(!fmt) {
      if(_debug_hdf)
	 fprintf(stderr,"When setting a node value, fmt has to be recognized\n");
      return 0;
   }

   if(!(node->data=malloc(sz))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate memory when setting value\n");
      return 0;
   }
   memcpy(node->data,value,sz);
   node->dSize=sz;

   if((strcmp(fmt,"string")==0) && typid<0) {
      node->typeId=createStringType(sz);
      if(node->typeId<0) {
	 if(_debug_hdf)
	    fprintf(stderr,"Failed to create string type\n");
	 return 0;
      }
   } else {
      strcpy(node->format,fmt);
      /*if(typid>=0)
	node->typeId=H5Tcopy(typid);*/
      if(typid>=0)
	 node->typeId=typid;
   }

   node->dataType=HL_SIMPLE;

   if(node->mark != NMARK_CREATED)
      node->mark=NMARK_CHANGED;

   return 1;
}

int setArrayValue(HL_Node* node,size_t sz,int ndims,hsize_t* dims,unsigned char* value,
		  const char* fmt,hid_t typid)
{
   int i;
   size_t npts=0;
   if(!fmt) {
      if(_debug_hdf)
	 fprintf(stderr,"When setting a node values, fmt has to be recognized\n");
      return 0;
   }

   npts=1;
   for(i=0;i<ndims;i++) {
      node->dims[i]=dims[i];
      npts*=dims[i];
   }
   if(!(node->data=malloc(npts*sz))) {
      if(_debug_hdf)
	 fprintf(stderr,"Failed to allocate memory when setting value\n");
      return 0;
   }
   memcpy(node->data,value,npts*sz);
   node->dSize=sz;
   node->ndims=ndims;
   strcpy(node->format,fmt);

   /*
    *if(typid>=0)
    *  node->typeId=H5Tcopy(typid);
    */
   if(typid>=0)
      node->typeId=typid;

   node->dataType=HL_ARRAY;

   if(node->mark != NMARK_CREATED)
      node->mark=NMARK_CHANGED;

   return 1;
}

void markAllNodes(HL_NodeList* nodelist, HL_NodeMark mark)
{
   int i;

   for(i=0;i<nodelist->nNodes;i++) {
      nodelist->nodes[i]->mark = mark;
   }
}

int extractParentChildName(HL_Node* node, char* parent, char* child)
{
   char tmpStr[256];
   char* tmpPtr;

   if(!parent || !child) {
      if(_debug_hdf)
	 fprintf(stderr,"When extracting ParentChild name, both parent and child must be specified\n");
      return 0;
   }

   strcpy(parent,"");
   strcpy(child,"");

   strcpy(tmpStr,node->name);
   if(!(tmpPtr=strrchr(tmpStr,'/'))) {
      if(_debug_hdf)
	 fprintf(stderr,"Could not extract '/' from node name %s\n",node->name);
      return 0;
   }
   tmpPtr[0]='\0';
   if(strcmp(tmpStr,"")!=0) {
      strcpy(parent,tmpStr);
   }
   tmpPtr++;
   strcpy(child,tmpPtr);
   return 1;
}

int commitDatatype(HL_Node* node,hid_t testStruct_hid)
{
   node->hdfId=testStruct_hid;
   return 1;
}

HL_CompoundTypeDescription* findCompoundTypeDescription(HL_NodeList* nodelist,
							    unsigned long objno0,
							    unsigned long objno1)
{
   int i;
   for(i=0;i<nodelist->nNodes;i++) {
      if(nodelist->nodes[i]->type==TYPE_ID && nodelist->nodes[i]->compoundDescription) {
	 if(objno0==nodelist->nodes[i]->compoundDescription->objno[0] &&
	    objno1==nodelist->nodes[i]->compoundDescription->objno[1])
	    return nodelist->nodes[i]->compoundDescription;
      }
   }
   return NULL;
}

void scanNodeList(HL_NodeList* nodelist)
{
   int i;

   for(i=0;i<nodelist->nNodes;i++) {
      printf("Name %s is ",nodelist->nodes[i]->name);
      if(nodelist->nodes[i]->type == ATTRIBUTE_ID)
	 printf("an attribute\n");
      else if(nodelist->nodes[i]->type == GROUP_ID)
	 printf("a group\n");
      else if(nodelist->nodes[i]->type == DATASET_ID)
	 printf("a dataset\n");
      else if(nodelist->nodes[i]->type == TYPE_ID)
	 printf("a datatype\n");
      else
	 printf("UNDEFINED\n");
   }
}

void scanCompoundTypeDescription(HL_CompoundTypeDescription* typelist)
{
   int i;
   if(!typelist) {
      if(_debug_hdf)
	 fprintf(stderr,"Can not scan NULL list\n");
      return;
   }
   printf("COMPOUND TYPE: '%s'\n",typelist->typename);
   for(i=0;i<typelist->nAttrs;i++) {
      printf("ATTRIBUTE '%s' format '%s' with rank '%d'\n",
	     typelist->attrs[i]->attrname, typelist->attrs[i]->format,
	     typelist->attrs[i]->ndims);
   }
}
